import logging
from typing import override

_ALLOWED_LOGGERS: set[str] = {"root"}


class HandlerRestrictedLogger(logging.Logger):
    @override
    def addHandler(self, hdlr: logging.Handler) -> None:
        if (
            isinstance(hdlr, logging.StreamHandler)
            and self.name not in _ALLOWED_LOGGERS
        ):
            return
        super().addHandler(hdlr)



import math
from collections.abc import Iterable

from liblaf.grapes.errors import UnreachableError

# multiplier, threshold, unit
SPECS: list[tuple[float, float, str]] = [
    (1e9, 1e3, "ns"),
    (1e6, 1e3, "µs"),
    (1e3, 1e3, "ms"),
    (1, 1e2, "s"),
]


def auto_duration_magnitude(seconds: float, *, significant: int = 3) -> int:
    number: float = float(f"{seconds:.{significant}e}")
    if number == 0:
        return -10
    seconds = round(seconds)
    if seconds < 100.0:
        return math.floor(math.log10(abs(number)))
    if seconds < 60 * 60:
        return 2
    if seconds < 24 * 60 * 60:
        return 3
    return 4


def pretty_duration(
    seconds: float, *, magnitude: int | None = None, significant: int = 3
) -> str:
    """.

    Examples:
        >>> pretty_duration(math.nan)
        '?? s'
        >>> pretty_duration(0)
        '.000 ns'
        >>> pretty_duration(1e-13)
        '.000 ns'
        >>> pretty_duration(1e-12)
        '.001 ns'
        >>> pretty_duration(1e-11)
        '.010 ns'
        >>> pretty_duration(1e-10)
        '.100 ns'
        >>> pretty_duration(1e-9)
        '1.00 ns'
        >>> pretty_duration(1e-8)
        '10.0 ns'
        >>> pretty_duration(1e-7)
        '100. ns'
        >>> pretty_duration(1e-6)
        '1.00 µs'
        >>> pretty_duration(1e-5)
        '10.0 µs'
        >>> pretty_duration(1e-4)
        '100. µs'
        >>> pretty_duration(1e-3)
        '1.00 ms'
        >>> pretty_duration(1e-2)
        '10.0 ms'
        >>> pretty_duration(1e-1)
        '100. ms'
        >>> pretty_duration(1.0)
        '1.00 s'
        >>> pretty_duration(1e1)
        '10.0 s'
        >>> pretty_duration(1e2)
        '01:40'
        >>> pretty_duration(1e3)
        '16:40'
        >>> pretty_duration(1e4)
        '02:46:40'
        >>> pretty_duration(1e5)
        '1d,03:46:40'
        >>> pretty_duration(1e6)
        '11d,13:46:40'
    """
    number: str
    unit: str
    number, unit = pretty_duration_unit(
        seconds, magnitude=magnitude, significant=significant
    )
    if unit:
        return f"{number} {unit}"
    return number


def pretty_duration_unit(
    seconds: float, *, magnitude: int | None = None, significant: int = 3
) -> tuple[str, str]:
    if not math.isfinite(seconds):
        return "??", "s"
    if seconds < 0:
        neg: str
        unit: str
        neg, unit = pretty_duration_unit(-seconds, significant=significant)
        return "-" + neg, unit
    if magnitude is None:
        magnitude = auto_duration_magnitude(seconds, significant=significant)
    magnitude = max(min(magnitude, 4), -10)
    for multiplier, threshold, unit in SPECS:
        number: float = seconds * multiplier
        precision: int = significant - magnitude - round(math.log10(multiplier)) - 1
        number = round(number, precision)
        if abs(number) < threshold:
            width: int = significant + 1
            precision = max(0, precision)
            number_formatted: str = f"{number:#.{precision}f}"
            if width == precision + 1:
                number_formatted = number_formatted.removeprefix("0")
            return number_formatted, unit
    seconds = round(seconds)
    match magnitude:
        case 2:
            minutes: int
            minutes, seconds = divmod(seconds, 60)
            return f"{minutes:02d}:{seconds:02d}", ""
        case 3:
            hours: int
            hours, seconds = divmod(seconds, 3600)
            minutes: int
            minutes, seconds = divmod(seconds, 60)
            return f"{hours:02d}:{minutes:02d}:{seconds:02d}", ""
        case 4:
            days: int
            days, seconds = divmod(seconds, 86400)
            hours: int
            hours, seconds = divmod(seconds, 3600)
            minutes: int
            minutes, seconds = divmod(seconds, 60)
            return f"{days:d}d,{hours:02d}:{minutes:02d}:{seconds:02d}", ""
    raise UnreachableError


def pretty_durations(
    seconds: Iterable[float],
    *,
    magnitude: int | None = None,
    significant: int = 3,
) -> tuple[list[str], str]:
    seconds = list(seconds)
    if magnitude is None:
        magnitude: int = auto_duration_magnitude(max(seconds), significant=significant)
    numbers: list[str] = []
    unit: str = ""
    for val in seconds:
        number: str
        number, unit = pretty_duration_unit(
            val, magnitude=magnitude, significant=significant
        )
        numbers.append(number)
    return numbers, unit



import functools
from collections.abc import Callable
from typing import Any


def wraps[C: Callable](wrapped: C) -> Callable[[Any], C]:
    def decorator(wrapper: Any) -> C:
        wrapper = functools.update_wrapper(wrapper, wrapped)
        return wrapper

    return decorator



from liblaf.grapes.conf import BaseConfig, Field, field


class ConfigTraceback(BaseConfig):
    indent_guide: Field[bool] = field(default=True)
    locals_hide_dunder: Field[bool] = field(default=True)
    locals_hide_sunder: Field[bool] = field(default=True)
    locals_max_length: Field[int] = field(default=10)
    locals_max_string: Field[int] = field(default=80)
    show_locals: Field[bool] = field(default=True)
    theme: Field[str] = field(default="ansi_dark")



from ._clear_handlers import clear_children_stream_handlers
from ._excepthook import install_excepthook
from ._handler import HandlerRestrictedLogger
from ._level import init_levels
from ._stacklevel import depth_logger
from ._unraisablehook import install_unraisablehook

__all__ = [
    "HandlerRestrictedLogger",
    "clear_children_stream_handlers",
    "depth_logger",
    "init_levels",
    "install_excepthook",
    "install_unraisablehook",
]



from ._by_name import FilterByName
from ._by_version import FilterByVersion
from ._composite import CompositeFilter
from ._factory import FilterLike, as_filter
from ._once import FilterOnce
from ._utils import as_levelno, as_levelno_dict

__all__ = [
    "CompositeFilter",
    "FilterByName",
    "FilterByVersion",
    "FilterLike",
    "FilterOnce",
    "as_filter",
    "as_levelno",
    "as_levelno_dict",
]



import logging

import attrs

from ._utils import as_levelno_dict


@attrs.define
class FilterByName:
    _levels: dict[str, int] = attrs.field(
        converter=as_levelno_dict, factory=lambda: {"__main__": logging.NOTSET}
    )

    def filter(self, record: logging.LogRecord) -> bool:
        level: int | None = self.get_level(record)
        if level is None:
            return True
        return record.levelno >= level

    def get_level(self, record: logging.LogRecord) -> int | None:
        name: str | None = record.name
        while True:
            if not name:
                return None
            level: int | None = self._levels.get(name)
            if level is not None:
                return level
            index: int = name.rfind(".")
            name = None if index < 0 else name[:index]



import inspect
import types
from collections.abc import Callable

from ._release_type import is_pre_release

_HIDDEN_FROM_LOGGING_NAMES: tuple[str, ...] = ("rich.progress",)


def hidden_from_logging(frame: types.FrameType | None) -> bool:
    if frame is None or frame.f_locals.get("__tracebackhide__", False):
        return True
    name: str = frame.f_globals.get("__name__", "")
    return name.startswith(_HIDDEN_FROM_LOGGING_NAMES)


def hidden_from_traceback(frame: types.FrameType | None) -> bool:
    if frame is None:
        return True
    if frame.f_locals.get("__tracebackhide__", False):
        return True
    name: str = frame.f_globals.get("__name__", "")
    return not is_pre_release(frame.f_code.co_filename, name)


def get_frame(
    depth: int = 1,
    hidden: Callable[[types.FrameType], bool] | None = hidden_from_traceback,
) -> types.FrameType | None:
    frame: types.FrameType | None = inspect.currentframe()
    while frame is not None and depth > 0:
        frame = frame.f_back
        depth -= 1
        if hidden is not None:
            while frame is not None and hidden(frame):
                frame = frame.f_back
    return frame



from . import custom
from ._auto import auto_pdoc
from ._options import make_kwargs
from ._pdoc import pdoc
from ._pformat import pformat
from ._pprint import pprint
from ._typing import CustomCallable, WadlerLindigOptions
from .custom import PdocCustomDispatcher, pdoc_custom, pdoc_fieldz, pdoc_rich_repr

__all__ = [
    "CustomCallable",
    "PdocCustomDispatcher",
    "WadlerLindigOptions",
    "auto_pdoc",
    "custom",
    "make_kwargs",
    "pdoc",
    "pdoc_custom",
    "pdoc_fieldz",
    "pdoc_rich_repr",
    "pformat",
    "pprint",
]



import contextlib
import contextvars
from collections.abc import Callable, Generator
from typing import Self, TypedDict, Unpack

import attrs
import wadler_lindig as wl
import wrapt
from rich.repr import RichReprResult

from liblaf.grapes.sentinel import MISSING

from ._environ import environ


@attrs.frozen
class Field[T]:
    _var: contextvars.ContextVar[T] = attrs.field(repr=False, alias="var")

    def __init__(
        self,
        name: str,
        default: T | MISSING = MISSING,
        *,
        env: str | None = None,
        factory: Callable[[], T] | None = None,
        type: type[T] | None = None,  # noqa: A002
        **_kwargs,
    ) -> None:
        if env is None:
            env = name.replace(".", "_").upper()
        value: T | MISSING = MISSING
        if value is MISSING and env is not None:
            value = environ.get(env, MISSING, typ=type)
        if value is MISSING and default is not MISSING:
            value = default
        if value is MISSING and factory is not None:
            value = factory()
        var: contextvars.ContextVar[T]
        if value is MISSING:
            var = contextvars.ContextVar(name)
        else:
            var = contextvars.ContextVar(name, default=value)
        self.__attrs_init__(var=var)  # pyright: ignore[reportAttributeAccessIssue]

    def __class_getitem__(cls, item: type[T]) -> type[Self]:
        return wrapt.partial(cls, type=item)  # pyright: ignore[reportReturnType]

    def __repr__(self) -> str:
        # TODO: use wadler-lindig
        return f"Field(name={self.name!r}, value={self.get()!r})"

    def __pdoc__(self, **kwargs) -> wl.AbstractDoc | None:
        pass

    def __rich_repr__(self) -> RichReprResult:
        yield "name", self.name
        yield "value", self.get()

    @property
    def name(self) -> str:
        return self._var.name

    def get(self) -> T:
        return self._var.get()

    def set(self, value: T, /) -> None:
        self._var.set(value)

    @contextlib.contextmanager
    def overrides(self, value: T, /) -> Generator[T]:
        token: contextvars.Token[T] = self._var.set(value)
        try:
            yield value
        finally:
            self._var.reset(token)


class FieldKwargs[T](TypedDict, total=False):
    default: T
    env: str
    factory: Callable[[], T]
    type: type[T]


def field[T](**kwargs: Unpack[FieldKwargs[T]]) -> Field[T]:
    return attrs.field(metadata=kwargs)



from typing import Any, dataclass_transform

import attrs

from ._define import define


@dataclass_transform(field_specifiers=(attrs.field,))
class AttrsMeta(type):
    def __new__[T: type](
        mcs: type[T],
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        /,
        **kwargs: Any,
    ) -> T:
        cls: T = super().__new__(mcs, name, bases, namespace)
        if "__attrs_attrs__" in namespace:
            return cls
        cls = define(cls, **kwargs)
        return cls


@dataclass_transform(field_specifiers=(attrs.field,))
class Attrs(metaclass=AttrsMeta): ...



import functools
from typing import Any

import attrs

from liblaf.grapes.functools import wraps
from liblaf.grapes.rich.repr import rich_repr_fieldz
from liblaf.grapes.wadler_lindig import pdoc_fieldz, pformat


@wraps(attrs.define)
def define(maybe_cls: type | None = None, **kwargs) -> Any:
    if maybe_cls is None:
        return functools.partial(define, **kwargs)
    auto_detect: bool = kwargs.get("auto_detect", True)
    repr_: bool | None = kwargs.get("repr")
    if auto_detect and repr_ is None and "__repr__" not in maybe_cls.__dict__:
        repr_ = True
    if repr_:
        maybe_cls.__repr__ = pformat  # pyright: ignore[reportAttributeAccessIssue]
        kwargs["repr"] = False
    if "__pdoc__" not in maybe_cls.__dict__:
        maybe_cls.__pdoc__ = pdoc_fieldz
    if "__rich_repr__" not in maybe_cls.__dict__:
        maybe_cls.__rich_repr__ = rich_repr_fieldz
    return attrs.define(maybe_cls, **kwargs)



from __future__ import annotations

from collections.abc import Container, Iterable
from typing import TYPE_CHECKING

from liblaf.grapes.logging import depth_logger

if TYPE_CHECKING:
    from _typeshed import SupportsContainsAndGetItem


_DEPRECATED_MESSAGE = "'%s' is deprecated. Please use '%s' instead."


def contains[T](
    obj: Container[T],
    key: T,
    deprecated_keys: Iterable[T] = (),
    *,
    msg: str = _DEPRECATED_MESSAGE,
) -> bool:
    if key in obj:
        return True
    for deprecated_key in deprecated_keys:
        if deprecated_key in obj:
            depth_logger.warning(msg, deprecated_key, key, stacklevel=2)
            return True
    return False


def getitem[KT, VT](
    obj: SupportsContainsAndGetItem[KT, VT],
    key: KT,
    deprecated_keys: Iterable[KT] = (),
    *,
    msg: str = _DEPRECATED_MESSAGE,
) -> object:
    if key in obj:
        return obj[key]
    for deprecated_key in deprecated_keys:
        if deprecated_key in obj:
            depth_logger.warning(msg, deprecated_key, key, stacklevel=2)
            return obj[deprecated_key]
    raise KeyError(key)



import logging

import attrs

from liblaf.grapes import magic

from ._utils import as_levelno


@attrs.define
class FilterByVersion:
    level_dev: int = attrs.field(default=logging.NOTSET, converter=as_levelno)
    level_pre: int = attrs.field(default=logging.DEBUG, converter=as_levelno)

    def filter(self, record: logging.LogRecord) -> bool:
        level: int | None = self.get_level(record)
        if level is None:
            return True
        return record.levelno >= level

    def get_level(self, record: logging.LogRecord) -> int | None:
        file: str = record.pathname
        name: str | None = record.name
        if magic.is_dev_release(file, name):
            return self.level_dev
        if magic.is_pre_release(file, name):
            return self.level_pre
        return None



import logging
from collections.abc import Mapping

import attrs
import cachetools

from ._by_name import FilterByName
from ._by_version import FilterByVersion
from ._once import FilterOnce
from ._utils import as_levelno


@attrs.define
class CompositeFilter:
    by_name: FilterByName = attrs.field(factory=FilterByName)
    by_version: FilterByVersion = attrs.field(factory=FilterByVersion)
    level: int = attrs.field(default=logging.WARNING)
    once: FilterOnce = attrs.field(factory=FilterOnce)

    _cache: cachetools.LRUCache[str, int] = attrs.field(
        repr=False, init=False, factory=lambda: cachetools.LRUCache(maxsize=1024)
    )

    def __init__(self, by_name: Mapping[str, int | str] | None = None) -> None:
        if by_name is None:
            by_name = {"__main__": logging.NOTSET}
        level: int = as_levelno(by_name.get("", logging.WARNING))
        self.__attrs_init__(by_name=FilterByName(by_name), level=level)  # pyright: ignore[reportAttributeAccessIssue]

    def filter(self, record: logging.LogRecord) -> bool:
        if not self.once(record):
            return False
        level: int = self.get_level(record)
        return record.levelno >= level

    def get_level(self, record: logging.LogRecord) -> int:
        cached: int = self._cache.get(record.name, -1)
        if cached == -1:
            level: int | None = self._get_level_uncached(record)
            if level is None:
                level = self.level
            self._cache[record.name] = cached = level
        return cached

    def _get_level_uncached(self, record: logging.LogRecord) -> int | None:
        level: int | None = self.by_name.get_level(record)
        if level is not None:
            return level
        level = self.by_version.get_level(record)
        if level is not None:
            return level
        return None



from ._ansi import has_ansi
from ._call import pretty_call
from ._duration import (
    auto_duration_magnitude,
    pretty_duration,
    pretty_duration_unit,
    pretty_durations,
)
from ._func import pretty_func
from ._throughput import pretty_throughput
from ._utils import get_name

__all__ = [
    "auto_duration_magnitude",
    "get_name",
    "has_ansi",
    "pretty_call",
    "pretty_duration",
    "pretty_duration_unit",
    "pretty_durations",
    "pretty_func",
    "pretty_throughput",
]



import types
import typing
from typing import Any, Protocol

import attrs
from environs import env

from liblaf.grapes.sentinel import MISSING


class EnvGetter(Protocol):
    def __call__(self, name: str, default: Any) -> Any: ...


@attrs.frozen
class Environ:
    _registry: dict[type, EnvGetter] = attrs.field(
        repr=False,
        init=False,
        factory=lambda: {
            int: env.int,
            bool: env.bool,
            str: env.str,
            list: env.list,
            dict: env.dict,
        },
    )

    def get[T](
        self, name: str, default: T | MISSING = MISSING, typ: type[T] | None = None
    ) -> T | MISSING:
        getter: EnvGetter = self._dispatch(typ)
        return getter(name, default)

    def _dispatch(self, typ: type | None) -> EnvGetter:
        if typ is None:
            return env
        getter: EnvGetter | None = self._registry.get(typ)
        if getter is not None:
            return getter
        if isinstance(typ, types.GenericAlias):
            return self._dispatch(typing.get_origin(typ))
        return env


environ = Environ()
